use std::{
    collections::HashMap,
    fs::{self, File},
    io::{self, Read},
    path::{Path, PathBuf},
};

use fluffle::{
    http::{
        response::{ResponseLine, Status},
        Request, Response,
    },
    HttpVersion, Server,
};

const PAGES_DIR: &str = "site/pages";
const TEMPLATES_DIR: &str = "site/templates";

fn walk_dir<P: AsRef<Path>>(dir: P, depth: i32) -> io::Result<Vec<PathBuf>> {
    let mut entries = Vec::new();

    if depth == 0 {
        return Ok(entries);
    }

    for entry in fs::read_dir(dir)? {
        let entry = entry?;
        let path = entry.path();

        entries.push(path.clone());

        if path.is_dir() {
            let sub_entries = if depth < 0 {
                walk_dir(&path, -1)?
            } else {
                walk_dir(&path, depth - 1)?
            };
            entries.extend(sub_entries);
        }
    }

    Ok(entries)
}

fn main() {
    let mut server = Server::new("127.0.0.1", 8080);

    server.on_request(|request: Request| -> Response {
        println!("{:#?}", request);

        let uri = request.request_line.target;
        let uri = uri.trim_start_matches("/");

        let uri = if uri.is_empty() { "main" } else { uri };

        for entry in walk_dir(PAGES_DIR, -1).unwrap() {
            let file_name = entry.file_name().and_then(|name| name.to_str()).unwrap();
            let file_stem = entry.file_stem().and_then(|stem| stem.to_str()).unwrap();
            let file_ext = entry.extension().and_then(|ext| ext.to_str()).unwrap();

            if file_name == uri || file_stem == uri {
                let mut file = File::open(&entry).unwrap();
                let mut content = String::new();
                file.read_to_string(&mut content).unwrap();

                let mut template_file = File::open(format!("{}/main.html", TEMPLATES_DIR)).unwrap();
                let mut template_content = String::new();
                template_file.read_to_string(&mut template_content).unwrap();

                if file_ext == "md" {
                    content = markdown::to_html(&content);
                }

                content = template_content.replace("{{content}}", &content);

                let response_line = ResponseLine::new(HttpVersion::default(), Status::OK);
                let content = Some(Vec::from(content));

                return Response {
                    response_line,
                    headers: None,
                    content,
                };
            }
        }

        let path = format!("site/{}", uri);
        let path = Path::new(&path);

        if path.exists() {
            let mut file = File::open(path).unwrap();
            let mut content = Vec::new();
            file.read_to_end(&mut content).unwrap();

            let file_ext = path.extension().and_then(|ext| ext.to_str()).unwrap();

            let response_line = ResponseLine::new(HttpVersion::default(), Status::OK);
            let headers = match file_ext {
                "js" => Some(HashMap::from([("Content-Type", "text/javascript")])),
                "wasm" => Some(HashMap::from([("Content-Type", "application/wasm")])),
                _ => None,
            };
            let content = Some(Vec::from(content));

            return Response {
                response_line,
                headers,
                content,
            };
        }

        if request.headers.get("Accept").unwrap().contains("text/html") {
            let mut file = File::open(format!("{}/error.html", TEMPLATES_DIR)).unwrap();
            let mut content = String::new();
            file.read_to_string(&mut content).unwrap();

            content = content.replace("{{code}}", "404");
            content = content.replace("{{message}}", "Not Found");

            let response_line = ResponseLine::new(HttpVersion::default(), Status::NotFound);
            let content = Some(Vec::from(content));

            return Response {
                response_line,
                headers: None,
                content,
            };
        }

        let response_line = ResponseLine::new(HttpVersion::default(), Status::NotFound);
        Response {
            response_line,
            headers: None,
            content: None,
        }
    });

    server.run();
}
